"""benchmark functions for optimization problems"""
import numpy as np


def sphere(x, none_variable=None):
    """
    The Sphere function has d local minima except for the
    global one. It is continuous, convex and unimodal.

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_SPHERE.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    n_dimensions = len(x)
    of = 0
    for i in range(n_dimensions):
        x_i = x[i]
        of += x_i ** 2

    return of


def rosenbrock(x, none_variable=None):
    """
    The Rosenbrock function is unimodal, and the global minimum 
    lies in a narrow, parabolic valley.
    
    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_ROSENBROCK.html

    Args:
        x (list): Design variables.

    Returns:
        of (float): Objective function value.
    """

    n_dimensions = len(x)
    sum = 0
    for i in range(n_dimensions - 1):
        x_i = x[i]
        x_next = x[i + 1]
        new = 100 * (x_next - x_i ** 2) ** 2 + (x_i - 1) ** 2
        sum += new
    of = sum

    return of


def rastrigin(x, none_variable=None):
    """
    The Rastrigin function has several local minima. It is highly multimodal, 
    but locations of the minima are regularly distributed.

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_RASTRIGIN.html

    Args:
        x (list): Design variables.

    Returns:
        of (float): Objective function value.
    """

    n_dimensions = len(x)
    sum = 0
    for i in range(n_dimensions):
        x_i = x[i]
        sum += (x_i ** 2 - 10 * np.cos(2 * np.pi * x_i))
    of = 10 * n_dimensions + sum

    return of


def ackley(x, none_variable=None):
    """
    The Ackley function in its two-dimensional form, it is characterized 
    by a nearly flat outer region, and a large hole at the centre.

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_ACKLEY.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    n_dimensions = len(x)
    sum1 = 0
    sum2 = 0
    a = 20
    b = 0.2
    c = 2 * np.pi
    for i in range(n_dimensions):
        x_i = x[i]
        sum1 += x_i ** 2
        sum2 += np.cos(c * x_i)
    term_1 = -a * np.exp(-b * np.sqrt(sum1 / n_dimensions))
    term_2 = -np.exp(sum2 / n_dimensions)
    of = term_1 + term_2 + a + np.exp(1)

    return of


def griewank(x, none_variable=None):
    """
    The Griewank function has many widespread local minima, 
    which are regularly distributed.

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_GRIEWANK.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    n_dimensions = len(x)
    sum = 0
    prod = 1
    for i in range(n_dimensions):
        x_i = x[i]
        sum += (x_i ** 2) / 4000
    prod *= np.cos(x_i / np.sqrt(i+1))
    of = sum - prod + 1

    return of


def zakharov(x, none_variable=None):
    """
    The Zakharov function has no local minima except the global one.

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_ZAKHAROV.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    n_dimensions = len(x)
    sum_1 = 0
    sum_2 = 0
    for i in range(n_dimensions):
        x_i = x[i]
        sum_1 += x_i ** 2
        sum_2 += (0.5 * i * x_i)
    of = sum_1 + sum_2**2 + sum_2**4

    return of


def easom(x, none_variable=None):
    """
    The Easom function has several local minima. It is unimodal, 
    and the global minimum has a small area relative to the search space.

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_EASOM.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    x_1 = x[0]
    x_2 = x[1]
    fact_1 = - np.cos(x_1) * np.cos(x_2)
    fact_2 = np.exp(- (x_1 - np.pi) ** 2 - (x_2 - np.pi) ** 2)
    of = fact_1*fact_2

    return of


def michalewicz(x, none_variable=None):
    """
    The Michalewicz function has d! local minima, and it is multimodal. 
    The parameter m defines the steepness of they valleys and ridges; 
    a larger m leads to a more difficult search.

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_MICHALEWICZ.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    n_dimensions = len(x)
    sum = 0
    m = 10
    for i in range(n_dimensions):
        x_i = x[i]
        sum += np.sin(x_i) * (np.sin((i * x_i ** 2) / np.pi)**(2 * m))
    of = -sum

    return of


def dixon_price(x, none_variable=None):

    """
    Dimensions: d

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_DIXONPRINCE.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    x1 = x[0]
    n_dimensions = len(x)
    term1 = (x1-1)**2
    sum = 0
    for i in range(1, n_dimensions):
        x_i = x[i]
        xold = x[i-1]
        new = i * (2*x_i**2 - xold)**2
        sum = sum + new
    of = term1 + sum

    return of


def goldstein_price(x, none_variable=None):
    """
    Dimensions: 2
    
    The Goldstein-Price function has several local minima.

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_GOLDSTAINPRICE.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    x1 = x[0]
    x2 = x[1]
    fact1A = (x1 + x2 + 1)**2
    fact1B = 19 - 14*x1 + 3*x1**2 - 14*x2 + 6*x1*x2 + 3*x2**2
    fact1 = 1 + fact1A * fact1B
    fact2A = (2*x1 - 3*x2)**2
    fact2B = 18 - 32*x1 + 12*x1**2 + 48*x2 - 36*x1*x2 + 27*x2**2
    fact2 = 30 + fact2A * fact2B
    of = fact1*fact2

    return of


def powell(x, none_variable=None):
    """
    Dimensions: d

    See documentation in https://wmpjrufg.github.io/METAPY/BENCH_POWELL.html

    Args:
        x (list): Design variables.
    
    Returns:
        of (float): Objective function value.
    """

    n_dimensions = len(x)
    sum = 0
    for i in range(1, n_dimensions//4 + 1):
        term1 = (x[4 * i - 3] + 10 * x[4 * i - 2])**2
        term2 = 5 * (x[4 * i-1] - x[4 * i])**2
        term3 = (x[4 * i - 2] - 2 * x[4 * i - 1])**4
        term4 = 10 * (x[4 * i - 3] - x[4 * i])**4
        sum = sum + term1 + term2 + term3 + term4
    of = sum

    return of


if __name__ == "__main__":

    #### Sphere function #####
    # Data
    x_sphere = [0, 0]

    # Call function
    of_sphere = sphere(x_sphere, None)

    # Output details
    print("of_best sphere: of = {:.4e}".format(of_sphere)) 

    #--------------------------------------------------------#
    
    ###### Rosenbrock function #####
    # Data
    x_rosenbrock = [1, 1]

    # Call function
    of_rosenbrock = rosenbrock(x_rosenbrock, None)

    # Output details
    print("of_best rosenbrock: of = {:.4f}".format(of_rosenbrock))

    #---------------------------------------------------------------#

    ###### Rastrigin function ######
    # Data
    x_rastrigin = [0, 0]

    # Call function
    of_rastrigin = rastrigin(x_rastrigin, None)

    # Output details
    print("of_best rastrigin: of = {:.4f}".format(of_rastrigin))

    #---------------------------------------------------------------#
    
    ###### Ackley function ######
    # Data
    x_ackley = [0, 0]

    # Call function
    of_ackley = ackley(x_ackley, None)

    # Output details
    print("of_best ackley: of = {:.4f}".format(of_ackley))

    #---------------------------------------------------------------#
    
    ###### Griewank function #####
    # Data
    x_griewank = [0, 0]

    # Call function
    of_griewank = griewank(x_griewank, None)

    # Output details
    print("of_best griewank: of = {:.4f}".format(of_griewank))

    #---------------------------------------------------------------#

    ###### Zakharov function #####
    # Data
    x_zakharov = [0, 0]

    # Call function
    of_zakharov = zakharov(x_zakharov, None)

    # Output details
    print("of_best zakharov: of = {:.4f}".format(of_zakharov))

    #---------------------------------------------------------------#

    ##### Easom function #####
    # Data
    x_easom = [0, 0]

    # Call function
    of_easom = easom(x_easom, None)

    # Output details
    print("of_best easom: of = {:.4f}".format(of_easom))

    #---------------------------------------------------------------#

    ##### Michalewicz function #####
    # Data
    x_michalewicz = [2.20, 1.57]

    # Call function
    of_michalewicz = michalewicz(x_michalewicz, None)

    # Output details
    print("of_best michalewicz: of = {:.4f}".format(of_michalewicz))

    #---------------------------------------------------------------#

    ##### DIXON-PRICE FUNCTION #####
    # Data
    i = 1
    numerator = 2**i - 2
    denominator = 2**i
    x_0 = 2**(-numerator/denominator)

    i = 2
    numerator = 2**i - 2
    denominator = 2**i
    x_1 = 2**(-numerator/denominator)
    x_dixonPrice = [x_0, x_1]

    # Call function
    of_dixonPrice = dixon_price(x_dixonPrice, None)

    # Output details
    print("of_best dixon-Price: of = {:.4f}".format(of_dixonPrice))

    #---------------------------------------------------------------#

    ##### Goldstein-Price function #####
    # Data
    x_goldsteinPrice = [0, -1]

    # Call function
    of_goldsteinPrice = goldstein_price(x_goldsteinPrice, None)

    # Output details
    print("of_best goldstein-Price: of = {:.4f}".format(of_goldsteinPrice))

    #---------------------------------------------------------------#

    ##### Powell function #####
    # Data
    x_powell = [0, 0]

    # Call function
    of_powell = powell(x_powell, None)

    # Output details
    print("of_best powell: of = {:.4f}".format(of_powell))

    #---------------------------------------------------------------#
